//---------------------------------------------------------------------------
#include <vcl.h>
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <malloc.h>
#pragma hdrstop

#include "tex.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------


typedef float MATRIX[4][4]; // matice

typedef struct {
    float x, y, z;
} VECTOR; // vektor

typedef struct {
    float x, y, z, u, v;
} VERTEX; // vrchol - bod

typedef struct {
    float a, b, c, d;
} PLANE; //  rovina

typedef struct {
    VERTEX *vertex[3];
    PLANE normal;
} FACE; // face = polygon

typedef struct {
    VERTEX *vertex;
    int vertex_num;
    FACE *face;
    int face_num;
    MATRIX matrix;
} OBJECT; // objekt

typedef struct {
    unsigned long pixel[65536]; // 256*256
} TEXTURE;

#define PI ((float)3.141592653589793238462643383279504228)
// Pi

// ----------------------------------------------- definice vjeci

#define LEN(A) ((float)sqrt(A.x*A.x+A.y*A.y+A.z*A.z))
// delka vektoru   MATH.h !! (sqrt)

#define NORMALIZE(A) {float s=LEN(A); A.x/=s;A.y/=s;A.z/=s;}
// normalizace vektoru (len je potom 1)

#define CROSS(A,B,C) C.x = B.y*A.z-A.y*B.z; \
                     C.y = B.z*A.x-A.z*B.x; \
                     C.z = B.x*A.y-A.x*B.y;
// vektorovy soucin

PLANE MakePl(FACE *poly)   // spocita rovinu polygonu poly
{
	PLANE p;
	VERTEX u, v, n;

	u.x = poly->vertex[1]->x - poly->vertex[0]->x;
	u.y = poly->vertex[1]->y - poly->vertex[0]->y;
	u.z = poly->vertex[1]->z - poly->vertex[0]->z;

	v.x = poly->vertex[2]->x - poly->vertex[0]->x;
	v.y = poly->vertex[2]->y - poly->vertex[0]->y;
	v.z = poly->vertex[2]->z - poly->vertex[0]->z;
    // spocita vektory hran polygonu

	CROSS(u,v,n); 
	NORMALIZE(n);
    // spocita vektor kolmy na rovinu polygonu

	p.a = n.x; 
	p.b = n.y; 
	p.c = n.z;
	p.d = -(n.x * poly->vertex[0]->x+
			n.y * poly->vertex[0]->y+
			n.z * poly->vertex[0]->z);
    // dopocita kolmou vzdalenost od pocatku [0,0,0] pomoci rovnice
    // roviny : "ax + by + cz + d = 0 <=> bod [x,y,z] lezi na rovine"

	return p;
}

// ----------------------------------------------- vektory

void initmatrix(MATRIX * m)
{
	int i,j;
	for (j=0; j<4; j++) for (i=0; i<4; i++) (*m)[i][j] = (i==j);
}

void matmult(MATRIX * a, MATRIX * b, MATRIX * r)
{
	int i, j;

	MATRIX tmp;
	for (i=0; i<4; i++)
	{
		for (j=0; j<4; j++)
		tmp[i][j] = (*a)[0][j] * (*b)[i][0] + (*a)[1][j] * (*b)[i][1] +
					 (*a)[2][j] * (*b)[i][2] + (*a)[3][j] * (*b)[i][3];
	}
	memcpy (r,tmp,sizeof(MATRIX));
}

void rotatematrixx(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[2][1] = (float)-sin(angle);
	temp[2][2] = (float)cos(angle);
	temp[1][1] = (float)cos(angle);
	temp[1][2] = (float)sin(angle);
	matmult(m,&temp,m);
}

void rotatematrixy(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[0][0] = (float)cos(angle);
	temp[0][2] = (float)-sin(angle);
	temp[2][2] = (float)cos(angle);
	temp[2][0] = (float)sin(angle);
	matmult(m,&temp,m);
}

void rotatematrixz(MATRIX * m, float angle)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[0][0] = (float)cos(angle);
	temp[0][1] = (float)sin(angle);
	temp[1][1] = (float)cos(angle);
	temp[1][0] = (float)-sin(angle);
	matmult(m,&temp,m);
}

void translatematrix(MATRIX * m, float dx, float dy, float dz)
{
	MATRIX temp;
	initmatrix(&temp);
	temp[3][0] = (float)dx;
	temp[3][1] = (float)dy;
	temp[3][2] = (float)dz;
	matmult(m,&temp,m);
}

void invertmatrix(MATRIX * src, MATRIX * inv)
{
	int i,j;
	float det, jdet;
	(*inv)[0][0] =   (*src)[1][1] * (*src)[2][2] - (*src)[1][2] * (*src)[2][1];
	(*inv)[0][1] = - (*src)[0][1] * (*src)[2][2] + (*src)[0][2] * (*src)[2][1];
	(*inv)[0][2] =   (*src)[0][1] * (*src)[1][2] - (*src)[0][2] * (*src)[1][1];
	det = (*inv)[0][0] * (*src)[0][0] + (*inv)[0][1] * (*src)[1][0] + (*inv)[0][2] * (*src)[2][0];
	if (det==0)
		return;
	(*inv)[1][0] = - (*src)[1][0] * (*src)[2][2] + (*src)[1][2] * (*src)[2][0];
	(*inv)[1][1] =   (*src)[0][0] * (*src)[2][2] - (*src)[0][2] * (*src)[2][0];
	(*inv)[1][2] = - (*src)[0][0] * (*src)[1][2] + (*src)[0][2] * (*src)[1][0];
	(*inv)[2][0] =   (*src)[1][0] * (*src)[2][1] - (*src)[1][1] * (*src)[2][0];
	(*inv)[2][1] = - (*src)[0][0] * (*src)[2][1] + (*src)[0][1] * (*src)[2][0];
	(*inv)[2][2] =   (*src)[0][0] * (*src)[1][1] - (*src)[0][1] * (*src)[1][0];
	jdet = 1 / det;
	for (j=0; j<3; j++) for (i=0; i<3; i++) (*inv)[j][i] *= jdet;
	(*inv)[3][0] = - (*src)[3][0]*(*inv)[0][0] - (*src)[3][1]*(*inv)[1][0] - (*src)[3][2]*(*inv)[2][0];
	(*inv)[3][1] = - (*src)[3][0]*(*inv)[0][1] - (*src)[3][1]*(*inv)[1][1] - (*src)[3][2]*(*inv)[2][1];
	(*inv)[3][2] = - (*src)[3][0]*(*inv)[0][2] - (*src)[3][1]*(*inv)[1][2] - (*src)[3][2]*(*inv)[2][2];
	(*inv)[0][3] = 0;
	(*inv)[1][3] = 0;
	(*inv)[2][3] = 0;
	(*inv)[3][3] = 1;
}

// ----------------------------------------------- matice

VERTEX __stdcall Mk_Vert(float x, float y, float z, float u, float v)
{
    VERTEX tmp;

    tmp.x = x;
    tmp.y = y;
    tmp.z = z;
    tmp.u = u;
    tmp.v = v;

    return tmp;
}

OBJECT *__stdcall Make_Cube(float size)
{
    OBJECT *tmp;
    VERTEX *v, center;
    FACE *f;
    int i;

    tmp = (OBJECT*) malloc(sizeof(OBJECT));
    tmp->face_num = 12;
    tmp->face = (FACE*) malloc(12 * sizeof(FACE));
    tmp->vertex_num = 24;
    tmp->vertex = (VERTEX*) malloc(24 * sizeof(VERTEX));
    // alokace ram

    v = tmp->vertex;

    v[0] = Mk_Vert(1, 1, 1, 0, 0);
    v[8] = Mk_Vert(1, 1, 1, 0, 255);
    v[12] = Mk_Vert(1, 1, 1, 0, 0);     //ok

    v[1] = Mk_Vert(1, 1, -1, 255, 0);
    v[9] = Mk_Vert(1, 1, -1, 255, 255);
    v[6] = Mk_Vert(1, 1, -1, 0, 0);       //ok

    v[2] = Mk_Vert(1, -1, 1, 0, 255);
    v[14] = Mk_Vert(1, -1, 1, 255, 0);
    v[20] = Mk_Vert(1, -1, 1, 0, 0);       //ok

    v[3] = Mk_Vert(1, -1, -1, 255, 255);
    v[4] = Mk_Vert(1, -1, -1, 0, 255);
    v[21] = Mk_Vert(1, -1, -1, 0, 255);     //ok

    v[5] = Mk_Vert(-1, -1, -1, 255, 255);
    v[17] = Mk_Vert(-1, -1, -1, 255, 255);
    v[22] = Mk_Vert(-1, -1, -1, 255, 255);   //ok

    v[15] = Mk_Vert(-1, -1, 1, 255, 255);
    v[16] = Mk_Vert(-1, -1, 1, 0, 255);
    v[23] = Mk_Vert(-1, -1, 1, 255, 0);      //ok

    v[11] = Mk_Vert(-1, 1, 1, 0, 0);
    v[13] = Mk_Vert(-1, 1, 1, 255, 0);
    v[19] = Mk_Vert(-1, 1, 1, 0, 0);         //ok
    
    v[7] = Mk_Vert(-1, 1, -1, 255, 0);
    v[10] = Mk_Vert(-1, 1, -1, 255, 0);
    v[18] = Mk_Vert(-1, 1, -1, 255, 0);       //ok

    center.x = 0;
    center.y = 0;
    center.z = 0;

    for(i = 0; i < tmp->vertex_num; i ++) {
        tmp->vertex[i].x *= size / 2;
        tmp->vertex[i].y *= size / 2;
        tmp->vertex[i].z *= size / 2;  // vertexy jsou v rozmezi +-1 (=> "2")

        center.x += tmp->vertex[i].x;
        center.y += tmp->vertex[i].y;
        center.z += tmp->vertex[i].z;
    }
    // nazvetsuje krychli a spocita stred

    center.x /= tmp->vertex_num;
    center.y /= tmp->vertex_num;
    center.z /= tmp->vertex_num;

    f = tmp->face;

    f[0].vertex[0] = &v[0];
    f[0].vertex[1] = &v[2];
    f[0].vertex[2] = &v[3];

    f[1].vertex[0] = &v[0];
    f[1].vertex[1] = &v[1];
    f[1].vertex[2] = &v[3];

    f[2].vertex[0] = &v[4];
    f[2].vertex[1] = &v[6];
    f[2].vertex[2] = &v[7];

    f[3].vertex[0] = &v[4];
    f[3].vertex[1] = &v[7];
    f[3].vertex[2] = &v[5];

    f[4].vertex[0] = &v[8];
    f[4].vertex[1] = &v[11];
    f[4].vertex[2] = &v[10];

    f[5].vertex[0] = &v[8];
    f[5].vertex[1] = &v[10];
    f[5].vertex[2] = &v[9];

    f[6].vertex[0] = &v[14];
    f[6].vertex[1] = &v[13];
    f[6].vertex[2] = &v[12];

    f[7].vertex[0] = &v[14];
    f[7].vertex[1] = &v[15];
    f[7].vertex[2] = &v[13];

    f[8].vertex[0] = &v[17];
    f[8].vertex[1] = &v[18];
    f[8].vertex[2] = &v[19];

    f[9].vertex[0] = &v[17];
    f[9].vertex[1] = &v[19];
    f[9].vertex[2] = &v[16];

    f[10].vertex[0] = &v[20];
    f[10].vertex[1] = &v[21];
    f[10].vertex[2] = &v[22];

    f[11].vertex[0] = &v[22];
    f[11].vertex[1] = &v[23];
    f[11].vertex[2] = &v[20];

    for(i = 0; i < tmp->face_num; i ++) {
        tmp->face[i].normal = MakePl(&tmp->face[i]);

        if(center.x * tmp->face[i].normal.a +
           center.y * tmp->face[i].normal.b +
           center.z * tmp->face[i].normal.c +
           tmp->face[i].normal.d < 0) {
            v = tmp->face[i].vertex[0];
            tmp->face[i].vertex[0] = tmp->face[i].vertex[1];
            tmp->face[i].vertex[1] = v;
            // prohodi dva vrcholy a otoci tak normalu
            tmp->face[i].normal = MakePl(&tmp->face[i]);
            // znova vygeneruje normalu
        }
    }
    // spocita normaly facu a otoci je smerem z objektu

    initmatrix(&tmp->matrix);
    // inicializuje matici objektu

    return tmp;
}

OBJECT *krychle;
MATRIX cmatrix;
TEXTURE textura;

// ------------------------------------------- konstrukce objektu

// nahraje texturu z kanvasu
void __stdcall Load_Tex_Canvas()
{
    long pix;
    int x, y;

    for(x = 0; x < 256; x ++) {
        for(y = 0; y < 256; y ++) {
            pix = Form1->bmp->Canvas->Pixels[x][y];
            // prevod RGB na GBR (quli TBitmap)
            pix = (pix & 0xff00) + ((pix >> 16) & 0xff) + ((pix & 0xff) << 16);
            //    green = green    red = blue             blue = red
            textura.pixel[x + 256 * y] = pix;
        }
    }
}

// ------------------------------------------- textura - nahravani

typedef struct {
    int x, y; // zatek hrany
    int num;  // poet dek (svisl dlka hrany)
    int dxdy; // delta x / delta y (o kolik se zmn x, jestlie se y zmn o 1)
} EDGE;

EDGE left_edge_buf[7], right_edge_buf[7],
    *left_edge, *right_edge;
int left_edge_num, right_edge_num;
// pr globlnch promnnch pro uchovvn hran

int Add_Edge(FACE *pgn, int actual, int next, EDGE *edge)
{
    float first_y, last_y;
    float x1, y1, x2, y2;
    float dxdy;
    int num;

    x1 = pgn->vertex[actual]->x;
    y1 = pgn->vertex[actual]->y;
    x2 = pgn->vertex[next]->x;
    y2 = pgn->vertex[next]->y;
    // souadnice zatku a konce hrany

    first_y = (float)ceil(y1);
    last_y  = (float)ceil(y2);
    num = (int)(last_y - first_y);
    // poet dek

    if(num == 0)
        return 0;
    // pokud je edge rovn, stejn ji pekryj jej sousedn edge
        
    dxdy  = (x2 - x1) / (y2 - y1);
    x1   += (first_y - y1) * dxdy;
    x1   *= 0x10000;
    dxdy *= 0x10000;
    // fixed point 16:16

    edge->dxdy = (int)dxdy;
    edge->x = (int)x1;
    edge->y = (int)first_y;
    edge->num = num;

    return 1;
}

int Create_Edges(FACE *pgn)
{
    int start_y,stol_y;
    int actual, top, bottom, next;
    float y1;
    int left_edge_start = 3 - 1;  // pocet vrcholu - 1
    int right_edge_stop = 0;
    float min_y =  0x10000;
    float max_y = -0x10000;
    
    left_edge_num = 0;
    right_edge_num = 0;
    // nastav poty hran na nulu

    actual = 0;
    while (actual < 3) { // 3 je pocet vrcholu face, funguje to i pro
        y1 = pgn->vertex[actual]->y;  // vicevrcholove KONVEXNI polygony
        if (y1 < min_y) {
            min_y = y1;
            top = actual;
        }
        if (y1 > max_y) {
            max_y = y1;
            bottom = actual;
        }
        actual ++;
    }
    start_y = (int)ceil(min_y);
    stol_y = (int)ceil(max_y) - 1;
    // hled se vrchol s min. a max. y
    
    // pokud by polygon nezabral ani jednu dku, nen vidt
    if(start_y > stol_y)
        return 0;
    
    // generovn prav vtve
    actual = top;
    while (actual != bottom) {
        next = actual + 1; 
        if(next >= 3)
            next = 0;
        if(Add_Edge(pgn,actual,next,&right_edge_buf[right_edge_stop])) {
            right_edge_stop ++;
            right_edge_num ++;
        } 
        actual = next;
    }
    
    // generovn lev vtve
    actual = bottom;
    while (actual != top) {
        next = actual + 1; 
        if(next >= 3)
            next = 0;
        if(Add_Edge(pgn,next,actual,&left_edge_buf[left_edge_start])) {
            left_edge_start --;
            left_edge_num ++;
        } 
        actual = next;
    }
    
    // nastav pointery na prvn hranu
    left_edge = left_edge_buf + left_edge_start + 1;
    right_edge = right_edge_buf;

    return 1;
}

// --------------------------------------------- hrany polygonu

float dudx, dudy, begu;
float dvdx, dvdy, begv;
float dzdx, dzdy, begz; // interpolacni konstanty

// spocita perspektivne zkreslene (hyperbola 1/x) mapovaci konstanty
int __stdcall Compute_Interpolation_Consts(FACE *poly)
{
    float x[2], y[2], z[2], u[2], v[2];
    float a, b, c;
    int i;

    x[0] = poly->vertex[1]->x - poly->vertex[0]->x;
    x[1] = poly->vertex[2]->x - poly->vertex[0]->x;

    y[0] = poly->vertex[1]->y - poly->vertex[0]->y;
    y[1] = poly->vertex[2]->y - poly->vertex[0]->y;
    
    z[0] = 1 / poly->vertex[1]->z - 1 / poly->vertex[0]->z;
    z[1] = 1 / poly->vertex[2]->z - 1 / poly->vertex[0]->z;
    // vsimnete si, ze se pouziva 1/z a ne z

    u[0] = 1 / poly->vertex[1]->z * poly->vertex[1]->u - 1 / poly->vertex[0]->z * poly->vertex[0]->u;
    u[1] = 1 / poly->vertex[2]->z * poly->vertex[2]->u - 1 / poly->vertex[0]->z * poly->vertex[0]->u;

    v[0] = 1 / poly->vertex[1]->z * poly->vertex[1]->v - 1 / poly->vertex[0]->z * poly->vertex[0]->v;
    v[1] = 1 / poly->vertex[2]->z * poly->vertex[2]->v - 1 / poly->vertex[0]->z * poly->vertex[0]->v;
    // vektory (od vrocholu 0 do 1, 2 pro kazdou souradnici)
    // vektory u a v jsou jeste nasobene z

    a = y[0] * z[1] - y[1] * z[0];
    b = z[0] * x[1] - z[1] * x[0];
    c = x[0] * y[1] - x[1] * y[0];
    // cross product - normala polygonu

    if(c == 0) 
        return 0;
    
    dzdx = - a / c;
    dzdy = - b / c;
    begz = (1 / poly->vertex[0]->z) - poly->vertex[0]->x * dzdx - poly->vertex[0]->y * dzdy;
    // Z

    a = y[0] * u[1] - y[1] * u[0];
    b = u[0] * x[1] - u[1] * x[0];
    //c = x[0] * y[1] - x[1] * y[0]; // c-cko je porad stejne
    // cross product - normala polygonu

    //if(c == 0) 
    //    return 0;
    
    dudx = - a / c;
    dudy = - b / c;
    begu = (poly->vertex[0]->u * 1 / poly->vertex[0]->z) - poly->vertex[0]->x *
        dudx - poly->vertex[0]->y * dudy;
    // U

    a = y[0] * v[1] - y[1] * v[0];
    b = v[0] * x[1] - v[1] * x[0];
    //c = x[0] * y[1] - x[1] * y[0]; // c-cko je porad stejne
    // cross product - normala polygonu

    //if(c == 0) 
    //    return 0;
    
    dvdx = - a / c;
    dvdy = - b / c;
    begv = (poly->vertex[0]->v * 1 / poly->vertex[0]->z) - poly->vertex[0]->x *
        dvdx - poly->vertex[0]->y * dvdy;
    // V

    return 1;
}

#define PIX_WEIGHT(l,w) (((((l & 0xff0000) * w) & 0xff000000) + \
                          (((l & 0x00ff00) * w) & 0x00ff0000) + \
                          (((l & 0x0000ff) * w) & 0x0000ff00)) >> 8)

void Interpolate_Segment(int u1, int du, int v1, int dv,
	unsigned long *video, unsigned long *v2)
{
    int w_1, w_2, w_3, w_4;
    long p1, p2, p3, p4;
    int uf, vf;

    do {
        uf = (u1 >> 12) & 0xf;
        vf = (v1 >> 12) & 0xf;

        w_1 = (16 - uf) * (16 - vf);
        w_3 = (16 - uf) * vf;
        w_2 = uf * (16 - vf);
        w_4 = uf * vf;

        //w_2 = 256 - (w_1 + w_2 + w_3);

        p1 = textura.pixel[(  (u1 >> 16)      & 0xff) + (  (v1 >> 8)      & 0xff00)];
        p2 = textura.pixel[( ((u1 >> 16) + 1) & 0xff) + (  (v1 >> 8)      & 0xff00)];
        p3 = textura.pixel[(  (u1 >> 16)      & 0xff) + ( ((v1 >> 8) + 0x100) & 0xff00)];
        p4 = textura.pixel[( ((u1 >> 16) + 1) & 0xff) + ( ((v1 >> 8) + 0x100) & 0xff00)];

        *video  = PIX_WEIGHT(p1,w_1);
        *video += PIX_WEIGHT(p2,w_2);
        *video += PIX_WEIGHT(p3,w_3);
        *video += PIX_WEIGHT(p4,w_4);

        video ++;
        // nova pozice

        u1 += du;
        v1 += dv;
        // interpolace souradnic
    } while(video < v2);
}

/*void Interpolate_Segment(int u1, int du, int v1, int dv,
	unsigned long *video, unsigned long *v2)   // No BILERP
{
    do {
        *video = textura.pixel[((u1 >> 16) & 0xff) + ((v1 >> 8) & 0xff00)];

        video ++;
        // nova pozice

        u1 += du;
        v1 += dv;
        // interpolace souradnic
    } while(video < v2);
}*/

void Draw_Segment(int c, int l, float z, float u, float v, unsigned long *video)
{
    int du, u1, u2;
    int dv, v1, v2;
    float z1;

    z1 = 0x10000 / z;
    // prevod zpatky na linearni, do 16:16 fp

    u1 = (int)(z1 * u);
    v1 = (int)(z1 * v);
    // prevod u, v probiha jinak..

    while(c >= 16) {
        z += dzdx * 16;
        u += dudx * 16;
        v += dvdx * 16;
        // pro optimalizaci prepocitame jen kazdy sestnacty pixel

        z1 = 0x10000 / z;
        // urceni dalsiho linearniho z

        u2 = (int)(z1 * u);
        v2 = (int)(z1 * v);
        // urceni dalsiho linearniho u, v

        du = (u2 - u1) / 16;
        dv = (v2 - v1) / 16;
        // spocita rovnici primky reprezentujici
        // vzdalenost polygonu od obrazovky

        Interpolate_Segment(u1, du, v1, dv, video, video + 16);
        // kreslici smycka

        c -= 16;
        // snizi pocet pixelu o 16

        video += 16;
        // posune ukazatele do pameti

        u1 = u2;
        v1 = v2;
        // presune se na dalsi souradnice
    }

    if(c > 0) { // kresli zbytek lajny (pokud delka neni nasobkem 16)
        z += dzdx * c;
        u += dudx * c;
        v += dvdx * c;

        z1 = 0x10000 / z;

        u2 = (int)(z1 * u);
        v2 = (int)(z1 * v);

        du = (u2 - u1) / c;
        dv = (v2 - v1) / c;

        Interpolate_Segment(u1, du, v1, dv, video, video + c);
        // kreslici smycka
    }
}

int DrawPolygon(FACE *p, unsigned long *video, TEXTURE *texture)
{
    int start_y, count, left_x, right_x,
        left_dxdy, right_dxdy, left_num,
        right_num, l, c;
    float z, u, v;

    if(!Create_Edges(p))
        return 0;

    if(!Compute_Interpolation_Consts(p))
        return 0;

    start_y = right_edge->y;
    count = right_edge_num;
    left_x = left_edge->x;
    left_dxdy = left_edge->dxdy;
    left_num = left_edge->num;
    right_x = right_edge->x;
    right_dxdy = right_edge->dxdy;
    right_num = right_edge->num;

    video += 320 * start_y;

    z = begz + dzdy * start_y;
    u = begu + dudy * start_y;
    v = begv + dvdy * start_y;
    // u, v, z prvniho radku rastru

    while(count > 0) {
        c = (left_x + 0x8000) / 0x10000;
        l = (right_x + 0x8000) / 0x10000;
        if(l < 0)
            l = 0;
        if(c >= 320)
            c = 319;
        c -= l ++;


        if(c > 0 && start_y >= 0 && start_y < 200) {
            //for(i = l; i < (l + c); i ++)
            //    video[i] = RGB(0, 0, 0);
            Draw_Segment(c, l, z + dzdx * l, u + dudx * l,
               v + dvdx * l, video + l);
        }
        // kresli vodorovny segment

        video += 320;
        start_y ++;
        z += dzdy;
        u += dudy;
        v += dvdy;
        // posun na dalsi radek obrazovky

        if(-- left_num == 0) {
            left_edge ++;
            left_x = left_edge->x;
            left_dxdy = left_edge->dxdy;
            left_num = left_edge->num;
        } else
            left_x += left_dxdy;
        if(-- right_num == 0) {
            count --;
            right_edge ++;
            right_x = right_edge->x;
            right_dxdy = right_edge->dxdy;
            right_num = right_edge->num;
        } else
            right_x += right_dxdy;
    }

    return 1;
}

// ------------------------------------------- texturovani

int mip = 1;

// nakresli objekt "obj" z pohledu "camera"
void DrawObject(MATRIX *camera, OBJECT *obj, unsigned long *videoram)
{
	float x, y, z, tx, ty, tz;
	VERTEX vtx[3];
	MATRIX m, im;
	int i, j;
	FACE fac;

	fac.vertex[0] = &vtx[0];
	fac.vertex[1] = &vtx[1];
	fac.vertex[2] = &vtx[2];
	// vertexy ve faci

	invertmatrix(camera, &m);
	matmult(&m, &obj->matrix, &m);
	invertmatrix(&m, &im);
	// transformacni matice

	for(i = 0; i < obj->face_num; i ++) {
        if(im[3][0] * obj->face[i].normal.a + im[3][1] * obj->face[i].normal.b +
           im[3][2] * obj->face[i].normal.c + obj->face[i].normal.d > 0.1)
            continue;
        // backface culling - odstrani neviditelne kurvy
		for(j = 0; j < 3; j ++) {
			x = obj->face[i].vertex[j]->x;
			y = obj->face[i].vertex[j]->y;
			z = obj->face[i].vertex[j]->z;

			tx = x * m[0][0] + y * m[1][0] + 
				 z * m[2][0] +     m[3][0];
			ty = x * m[0][1] + y * m[1][1] + 
				 z * m[2][1] +     m[3][1];
			tz = x * m[0][2] + y * m[1][2] + 
				 z * m[2][2] +     m[3][2];
			// transformace

			if(tz > 0.1){
				tx = (320 / 2) + (tx * 256) / tz;
				ty = (200 / 2) + (ty * 256) / tz;  // 256 je perspektivni skresleni
				fac.vertex[j]->x = tx;
				fac.vertex[j]->y = ty;
				fac.vertex[j]->z = tz;

                fac.vertex[j]->u = obj->face[i].vertex[j]->u / mip;
				fac.vertex[j]->v = obj->face[i].vertex[j]->v / mip;  // souradnice textury 0-255

                //if(tx >= 0 && tx <= 320/*scr.Width*/ && ty >= 0 && ty < 200/*scr.Height*/)
                //    videoram[(int)tx + 320 * (int)ty] = RGB(0, 0, 0);
			}
			// perspektivni korekce
		}
        
		DrawPolygon(&fac, videoram, &textura);
	}
}

// ------------------------------------------- kresleni

void __fastcall TForm1::FormCreate(TObject *Sender)
{
    krychle = Make_Cube(50);
    // udela kostku 50 * 50 * 50

    initmatrix(&cmatrix);
	//rotatematrixz(&cmatrix, (float)-PI);
	//rotatematrixy(&cmatrix, (float)-PI/6);
	translatematrix(&cmatrix, 0, 0, -120);
    // vytvori matici kamery

    Load_Tex_Canvas();
    // nahraje texturu z kanvasu bmp

    tmr_rotation->Enabled = True;

    //Form1->scr->Canvas->Pixels[10][10] = RGB(0, 0, 0);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::cmdDrawClick(TObject *Sender)
{
    if(tmr_rotation->Enabled == True)
        tmr_rotation->Enabled = False;
    else
        tmr_rotation->Enabled = True;
}
//---------------------------------------------------------------------------

long last_frame_render;
long frame_num;

void __fastcall TForm1::tmr_rotationTimer(TObject *Sender)
{
    long current_render;
    Graphics::TBitmap *bmp = new Graphics::TBitmap();
    unsigned long video[64000];   // 64000 = 320*200 - potrebuju pamet tohohle tvaru
    int x, y;
    float n_FPS;

    for(x = 0; x < 64000; x ++)
        video[x] = RGB(255, 255, 255);

    rotatematrixx(&krychle->matrix, PI / 2000);
    rotatematrixz(&krychle->matrix, -PI / 2000);
    //
    DrawObject(&cmatrix, krychle, video);

    bmp->Width = 320;
    bmp->Height = 200;
    bmp->PixelFormat = (Graphics::TPixelFormat)7;
    bmp->IgnorePalette = true;

    //for(x = 00; x < 320; x ++) {
    for(y = 0; y < 200; y ++)
        memcpy((long*)bmp->ScanLine[y], &video[320 * y], 320 * sizeof(long));
            //Form1/*->scr*/->Canvas->Pixels[x][y] = video[x + 320 * y];
    //}
    Form1->scr->Canvas->Draw(0, 0, bmp);

    delete bmp;

    current_render = clock();
    n_FPS = (float)frame_num / ((float)(current_render - last_frame_render) / (float)CLOCKS_PER_SEC);
    frame_num ++;
    if((current_render - last_frame_render) / (float)CLOCKS_PER_SEC >= 1) {
        frame_num = 0;
        last_frame_render = clock();

        char str[256];

        sprintf(str, "%.2f FPS", n_FPS);
        lbl_FPS->Caption = str;
    }
    // pocitani FPS, label se zupdatuje kazdou sekundu


   /* memcpy((long*)(Form1->scr->Canvas->Pixels), video, 64000 * sizeof(long));

    if(Form1->scr->Canvas->Pixels[10][10] == RGB(0, 0, 0))
        Form1->scr->Canvas->Pixels[10][10] = RGB(255, 255, 255);
    else
        Form1->scr->Canvas->Pixels[10][10] = RGB(0, 0, 0);
    // pomaleeee - potrebujes jiny rozhrani nez kanvas  */
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Mip_addClick(TObject *Sender)
{
    if(mip > 1)
        mip /= 2;
}
//---------------------------------------------------------------------------

void __fastcall TForm1::Mip_subClick(TObject *Sender)
{
    mip *= 2;
}
//---------------------------------------------------------------------------

